/*
 * backconn-adm
 *
 * This file is part of the backconn project. http://backconn.sourceforge.net
 * Copyright (C) 2017, 2018 Mateusz Viste
 *
 * ===================== PUBLISHED UNDER THE MIT LICENSE =====================
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * ===========================================================================
 */

#include <arpa/inet.h>
#include <ctype.h>       /* tolower() */
#include <errno.h>
#include <limits.h>      /* PATH_MAX */
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>      /* atoi() */
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>    /* umask() */
#include <sys/time.h>    /* struct timeval */
#include <sys/types.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include "config.h"
#include "common.h"      /* sendall() */


enum tflags {
  F_NOFORK = 1,
  F_NOBLOCK = 2,
  F_CFGRELOAD = 4
};

/* a few (nasty) global variables related to configuration */
struct scfg *cfg;
char cfgfname[PATH_MAX + 1];

/* reload configuration, if needed. returns 0 if no need, 1 if reloaded and
 * -1 on error */
int reloadcfgifneeded(void) {
  static time_t cfglastcheck = 0;
  static time_t cfglastmod = 0;
  struct stat stbuf;
  time_t curtime;
  FILE *cfgfp;

  curtime = time(NULL);
  /* detect time changes into the future */
  if (cfglastcheck > curtime) cfglastcheck = 0;
  /* if checked less than 10s ago, don't bother */
  if (cfglastcheck + 10 > curtime) return(0);
  cfglastcheck = curtime;
  /* if file's mtime changed, reload config */
  if (stat(cfgfname, &stbuf) != 0) return(-1);
  if (stbuf.st_mtime == cfglastmod) return(0);
  cfglastmod = stbuf.st_mtime;
  /* reload config */
  config_free(cfg);
  cfgfp = fopen(cfgfname, "rb");
  if (cfgfp == NULL) return(-1);
  cfg = config_read(cfgfp);
  fclose(cfgfp);
  return(1);
}


/* trims the '::ffff:' prefix from an IP string, if found */
static void trimipv6mapping(char *addr) {
  char *prefix = "::ffff:";
  int i;
  /* is the prefix present? */
  for (i = 0; prefix[i] != 0; i++) if (tolower(addr[i]) != prefix[i]) return;
  /* move addr 7 bytes to the left */
  memmove(addr, addr+7, strlen(addr) - 6);
}


/* Waits for a connection, forks when a client connection arrives, and
 * returns the forked socket. Fills clientipaddrstr and serveripaddrstr with
 * IP addresses (src and dst) */
static int waitforconn(int port, char *clientipaddrstr, int clientipaddrstr_maxlen, int flags) {
  int sockmaster, sockslave;
  int one = 1;  /* this is used by setsockopt() calls on the socket later */
  socklen_t clilen;
  struct sockaddr_in6 serv_addr, cli_addr;
  pid_t mypid;

  serv_addr.sin6_addr = in6addr_any;
  sockmaster = socket(AF_INET6, SOCK_STREAM, 0);
  if (sockmaster < 0) {
    syslog(LOG_WARNING, "FATAL ERROR: socket could not be open (%s)", strerror(errno));
    return(-2);
  }

  /* I set the socket to be reusable, to avoid having to wait for a longish time when the server is restarted */
  if (setsockopt(sockmaster, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one)) < 0) syslog(LOG_WARNING, "WARNING: failed to set REUSEADDR on main socket");

  /* Initialize socket structure */
  memset(&serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin6_family = AF_INET6;
  serv_addr.sin6_addr = in6addr_any;  /* bind all */
  serv_addr.sin6_port = htons(port);

  /* Explicitely mark the socket as NOT being IPV6-only. This is needed on systems that have a system-wide sysctl net.inet6.ip6.v6only=1 (by default every *nix besides Linux...) */
  #ifdef IPV6_BINDV6ONLY  /* check if the BINDV6ONLY option exists at all */
    {
      int zero = 0;
      setsockopt(sockmaster, IPPROTO_IPV6, IPV6_BINDV6ONLY, (char *)&zero, (socklen_t)sizeof(zero));
    }
  #endif

  /* Now bind the host address using a bind() call */
  if (bind(sockmaster, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
    syslog(LOG_WARNING, "FATAL ERROR: binding failed (%s)", strerror(errno));
    return(-2);
  }

  /* Start listening for clients */
  listen(sockmaster, 10);
  clilen = sizeof(cli_addr);

  /* Ignore SIGCHLD - this way I don't have to worry about my children becoming little zombies */
  signal(SIGCHLD, SIG_IGN);

  /* I don't want to get notified about SIGHUP */
  signal(SIGHUP, SIG_IGN);

  /* Set the user file creation mask to zero. */
  umask(0);

  /* Redirect standard file descriptors to /dev/null - no reason for a daemon to write to stdout, stderr or stdin */
  freopen("/dev/null", "r", stdin);
  freopen("/dev/null", "w", stdout);
  freopen("/dev/null", "w", stderr);

  /* I want to be the pack master now (aka session leader) */
  if (setsid() == -1) syslog(LOG_WARNING, "WARNING: setsid() failed (%s)", strerror(errno));

  /* set the working directory to the root directory */
  if (chdir ("/") == -1) syslog(LOG_WARNING, "WARNING: failed to switch to / directory (%s)", strerror(errno));

  /* if F_NOBLOCK, then stop here */
  if (flags & F_NOBLOCK) return(sockmaster);

  for (;;) {
    /* Accept actual connection from the client - here process will go to sleep mode, waiting for incoming connections */
    sockslave = accept(sockmaster, (struct sockaddr *)&cli_addr, &clilen);
    if (sockslave < 0) {
      syslog(LOG_WARNING, "FATAL ERROR: accepting connection failed (%s)", strerror(errno));
      close(sockmaster);
      return(-2);
    }

    /* reload configuration, if requested and if needed */
    if (flags & F_CFGRELOAD) {
      int r = reloadcfgifneeded();
      if (r < 0) {
        syslog(LOG_ERR, "ERROR: failed to reload configuration file '%s'!", cfgfname);
      } else if (r > 0) {
        syslog(LOG_ERR, "config file reloaded ('%s')", cfgfname);
      }
    }

    if (flags & F_NOFORK) {
      close(sockmaster);
      return(sockslave);
    }

    /* fork out, close the master socket and return the client socket */
    mypid = fork();
    if (mypid == 0) { /* I'm the child */
      static char logprefix[128];
      close(sockmaster);
      if (inet_ntop(cli_addr.sin6_family, &cli_addr.sin6_addr, clientipaddrstr, clientipaddrstr_maxlen) == NULL) {
        syslog(LOG_WARNING, "Failed to fetch client's IP address: %s", strerror(errno));
        sprintf(clientipaddrstr, "UNKNOWN");
      }
      /* process ipv6-mapped addresses so they look clean */
      trimipv6mapping(clientipaddrstr);
      /* set logprefix to contain the client's address */
      sprintf(logprefix, "backconn-adm [%s]", clientipaddrstr);
      openlog(logprefix, LOG_PID, LOG_DAEMON); /* set up the logging to log with PID and peer's IP address */
      syslog(LOG_INFO, "new connection");
      return(sockslave);
    } else if (mypid > 0) { /* I'm the parent */
      /* just close child's socket to avoid messing with it */
      close(sockslave);
    } else { /* error condition */
      syslog(LOG_WARNING, "FATAL ERROR: fork() failed!");
      close(sockslave);
      close(sockmaster);
      return(-2);
    }
  }
}


int main(int argc, char **argv) {
  int connremote, connadm;
  int res;
  int xorkey = 0;
  int listenport; /* global port backconn-adm listen to for clients */
  int lport;      /* local port for tunneled data */
  int rport;      /* remote port we want to reach on the client */
  char clientipaddrstr[128];
  char clientname[256];
  char adminipaddrstr[128];
  unsigned char buf[8192];
  struct scfg *cfgitem;
  pid_t mypid;

  if (argc != 3) {
    fprintf(stderr, "usage: backconn-adm config.conf listenport\n"
                    "\n"
                    "example: backconn-adm /etc/backconn.conf 9000\n");
    return(1);
  }
  if (realpath(argv[1], cfgfname) == NULL) {
    fprintf(stderr, "ERROR: failed to find the path to '%s')\n", argv[1]);
    return(1);
  }

  /* read listen port */
  listenport = atoi(argv[2]);
  if ((listenport > 65535) || (listenport < 1)) {
    fprintf(stderr, "ERROR: listen port out of range (should be between 1 and 65535, got '%s')\n", argv[2]);
    return(1);
  }
  /* read configuration */
  if (reloadcfgifneeded() < 0) {
    fprintf(stderr, "ERROR: failed to open configuration file '%s'\n", cfgfname);
    return(1);
  }
  if (cfg == NULL) fprintf(stderr, "WARNING: empty configuration file '%s'\n", cfgfname);

  syslog(LOG_INFO, "backconn-adm v" pVer " process started, will listen on port TCP/%d", listenport);
  printf("backconn-adm process started, will listen on port TCP/%d\n", listenport);

  /* fork off */
  mypid = fork();
  if (mypid == 0) { /* I'm the child, do nothing */
    /* nothing to do - just continue */
  } else if (mypid > 0) { /* I'm the parent - quit now */
    return(0);
  } else {  /* error condition */
    syslog(LOG_WARNING, "Failed to dameonize (%s)", strerror(errno));
    fprintf(stderr, "Failed to dameonize (%s)", strerror(errno));
    return(1);
  }

  /* wait for a remote connexion (and fork) */
  connremote = waitforconn(listenport, clientipaddrstr, sizeof(clientipaddrstr), F_CFGRELOAD);
  if (connremote < 0) return(0); /* if -1 then it's the forked parent */
  enable_tcpkeepalive(connremote);

  /* remote side should advertise a name */
  if (recvtimeout(connremote, buf, 1, 10, 0) != 1) {
    syslog(LOG_WARNING, "connection broken (line #%d)", __LINE__);
    close(connremote);
    return(1);
  }
  if (buf[0] == 0) {
    syslog(LOG_INFO, "remote host didn't advertise any name");
    clientname[0] = 0;
  } else {
    if (recvtimeout(connremote, clientname, buf[0], 10, MSG_WAITALL) != buf[0]) {
      syslog(LOG_WARNING, "connection broken (line #%d)", __LINE__);
      close(connremote);
      return(1);
    }
    clientname[buf[0]] = 0;
    syslog(LOG_INFO, "remote host advertised itself as '%s'", clientname);
  }

  /* match a configuration entry (first on IP address, then on clientname) */
  cfgitem = config_find(cfg, clientipaddrstr);
  if ((cfgitem == NULL) && (clientname[0] != 0)) cfgitem = config_find(cfg, clientname);
  if (cfgitem == NULL) {
    if (clientname[0] != 0) {
      syslog(LOG_ERR, "could not match any configuration entry to either '%s' nor '%s'", clientipaddrstr, clientname);
    } else {
      syslog(LOG_ERR, "could not match any configuration entry to '%s'", clientipaddrstr);
    }
    return(1);
  }
  rport = atoi(cfgitem->val[0]);
  lport = atoi(cfgitem->val[1]);
  if (cfgitem->val[2] != NULL) {
    xorkey = atoi(cfgitem->val[2]);
    if ((xorkey > 255) || (xorkey < 1)) {
      xorkey = 0;
      syslog(LOG_WARNING, "Warning: invalid xor key for %s (should be between 1 and 255, found '%s')", cfgitem->token, cfgitem->val[2]);
    }
  }
  if ((rport < 1) || (rport > 65535)) {
    syslog(LOG_ERR, "Error: invalid rport for %s (should be between 1 and 65535, found '%s')", cfgitem->token, cfgitem->val[0]);
    config_free(cfg);
    close(connremote);
    return(1);
  }
  if ((lport < 1) || (lport > 65535)) {
    syslog(LOG_ERR, "Error: invalid lport for %s (should be between 1 and 65535, found '%s')", cfgitem->token, cfgitem->val[1]);
    config_free(cfg);
    close(connremote);
    return(1);
  }

  /* free config */
  config_free(cfg);

  /* open an admin socket locally */
  syslog(LOG_INFO, "opening admin socket on TCP/%u for client %s:%d", lport, clientipaddrstr, rport);
  connadm = waitforconn(lport, adminipaddrstr, sizeof(adminipaddrstr), F_NOBLOCK);
  if (connadm < 0) {
    syslog(LOG_INFO, "opening admin socket failed: %s", strerror(errno));
    close(connremote);
    return(1);
  }
  /* wait for a client to come on the admin socket, while keeping the control channel alive at the same time */
  for (;;) {
    fd_set fds;
    struct timeval t;
    int highestfd, selectres;
    unsigned char bbuf[2];
    t.tv_sec = 20;  /* this controls the application-layer keepalive freq */
    t.tv_usec = 0;
    if (connadm > connremote) {
      highestfd = connadm;
    } else {
      highestfd = connremote;
    }
    FD_ZERO(&fds);
    FD_SET(connadm, &fds);
    FD_SET(connremote, &fds);
    selectres = select(highestfd+1, &fds, NULL, NULL, &t);
    if (selectres < 0) {
      syslog(LOG_ERR, "select() error");
      close(connremote);
      close(connadm);
      return(0);
    }
    if (FD_ISSET(connadm, &fds)) {
      socklen_t clilen;
      int newconn;
      struct sockaddr_in6 cli_addr;
      syslog(LOG_INFO, "got activity on adm sock");
      newconn = accept(connadm, (struct sockaddr *)&cli_addr, &clilen);
      close(connadm);
      connadm = newconn;
      break;
    }
    if (FD_ISSET(connremote, &fds)) {
      /* it is unexpected for the remote connection to send anything at this point - closing */
      syslog(LOG_WARNING, "received unexpected data or connection aborted");
      close(connremote);
      close(connadm);
      return(0);
    }
    /* send keepalive (0x00 0x00) and expect immediate answer (0xAA) */
    memset(bbuf, 0, 2);
    xorstr(bbuf, xorkey, 2);
    if ((sendall(connremote, bbuf, 2) != 0) || (recv(connremote, bbuf, 1, 0) != 1) || (bbuf[0] != (0xAA ^ xorkey))) {
      syslog(LOG_WARNING, "keepalive failure");
      close(connremote);
      close(connadm);
      return(0);
    }
  }
  /* enable tcp keepalive on the new connection */
  enable_tcpkeepalive(connadm);

  /* send request to remote client for opening port xxx */
  buf[0] = rport >> 8;
  buf[1] = rport & 0xff;
  xorstr(buf, xorkey, 2);
  if (sendall(connremote, buf, 2) != 0) {
    syslog(LOG_ERR, "send() of port string failed");
    close(connremote);
    close(connadm);
    return(0);
  }

  /* transfer data back and forth, until one of the connexions breaks */
  res = proxy(connremote, connadm, buf, sizeof(buf), xorkey);
  syslog(LOG_INFO, "proxy() result: %d", res);
  close(connremote);
  close(connadm);

  return(0);
}
