/*
 * backconn-cli
 *
 * This file is part of the backconn project. http://backconn.sourceforge.net
 * Copyright (C) 2017, 2018 Mateusz Viste
 *
 * ===================== PUBLISHED UNDER THE MIT LICENSE =====================
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * ===========================================================================
 */

#include <errno.h>
#include <signal.h>
#include <stdio.h>   /* printf() and friends */
#include <stdlib.h>  /* atoi() */
#include <unistd.h>  /* sleep() */
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <string.h>  /* memset() */

#include "common.h"      /* sendall() */


static int syslogflag = 0;  /* log to syslog if set (argument '-s') */
static int stderrflag = 1;  /* log errors to stderr */

/* parse and validates arguments. returns 0 on success, non-zero otherwise */
static int parseargv(int argc, char **argv, char **rhost, char **rport, int *xorkey, char **name, int *namelen, int *reconntime, int *daemonize) {
  int i, flags = 0;
  /* load defaults */
  *rhost = NULL;
  *rport = NULL;
  *name = NULL;
  *namelen = 0;
  *xorkey = 0;
  *reconntime = 5;
  *daemonize = 0;
  /* parse arguments */
  for (i = 1; i < argc; i++) {
    /* is it an option argument? */
    if (argv[i][0] == '-') {
      if ((argv[i][1] == 'x') && (argv[i][2] == 0) && (i+1 < argc)) {
        *xorkey = atoi(argv[++i]);
      } else if ((argv[i][1] == 't') && (argv[i][2] == 0) && (i+1 < argc)) {
        *reconntime = atoi(argv[++i]);
        if (*reconntime == 0) {
          fprintf(stderr, "Error: invalid -t value (must be greater than 0)\n");
          return(-1);
        }
      } else if ((argv[i][1] == 'n') && (argv[i][2] == 0) && (i+1 < argc)) {
        *name = argv[++i];
        *namelen = strlen(*name);
        if (*namelen > 255) {
          if (stderrflag != 0) fprintf(stderr, "Error: name too long\n");
          return(-1);
        }
      } else if ((argv[i][1] == 's') && (argv[i][2] == 0)) {
        syslogflag = 1;
      } else if ((argv[i][1] == 'q') && (argv[i][2] == 0)) {
        stderrflag = 0;
      } else if ((argv[i][1] == 'd') && (argv[i][2] == 0)) {
        *daemonize = 1;
      } else {
        fprintf(stderr, "Error: invalid option\n");
        return(-1);
      }
    } else if (*rhost == NULL) { /* if not having an IP yet, this would be it */
      *rhost = argv[i];
    } else if (*rport == NULL) { /* if not having a remote port yet, this is it */
      *rport = argv[i];
      if ((atoi(*rport) < 1) || (atoi(*rport) > 65535)) {
        fprintf(stderr, "Error: invalid rport\n");
        return(-1);
      }
    } else { /* no idea what this is */
      fprintf(stderr, "Error: invalid argument\n");
      return(-1);
    }
  }
  /* did I get at the very least a host and a port? */
  if ((*rhost == NULL) || (*rport == NULL)) {
    fprintf(stderr, "Error: missing arguments\n");
    return(-1);
  }
  return(flags);
}


/* establishes a connection to host and port. returns a valid socket on success, -1 on error. */
static int connect_to(char *host, char *port) {
  struct addrinfo hints, *result, *rp;
  int s, sfd;
  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_family = AF_UNSPEC;     /* Allow IPv4 or IPv6 */
  hints.ai_socktype = SOCK_STREAM; /* Datagram socket */
  hints.ai_flags = AI_PASSIVE | AI_NUMERICSERV;  /* For wildcard IP address */
  hints.ai_protocol = 0;           /* Any protocol */
  hints.ai_canonname = NULL;
  hints.ai_addr = NULL;
  hints.ai_next = NULL;

  s = getaddrinfo(host, port, &hints, &result);
  if (s != 0) {
    if (stderrflag != 0) fprintf(stderr, "getaddrinfo failure on '%s': %s\n", host, gai_strerror(s));
    if (syslogflag != 0) syslog(LOG_ERR, "getaddrinfo failure on '%s': %s\n", host, gai_strerror(s));
    return(1);
  }

  /* getaddrinfo() returns a list of address structures.
     Try each address until we successfully bind(2).
     If socket(2) (or bind(2)) fails, we (close the socket
     and) try the next address. */

  for (rp = result; rp != NULL; rp = rp->ai_next) {
    sfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
    if (sfd == -1) continue;
    /* connect now */
    if (connect(sfd, rp->ai_addr, sizeof(struct sockaddr)) < 0) {
      if (stderrflag != 0) fprintf(stderr, "Error: connect() to server %s:%s failed (%s)\n", host, port, strerror(errno));
    } else { /* success */
      break;
    }
  }

  if (rp == NULL) { /* No address succeeded */
    if (stderrflag != 0) fprintf(stderr, "connection failure\n");
    if (syslogflag != 0) syslog(LOG_ERR, "connection failure");
    freeaddrinfo(result);           /* No longer needed */
    return(-1);
  }

  freeaddrinfo(result);           /* No longer needed */

  enable_tcpkeepalive(sfd);
  return(sfd);
}


int main(int argc, char **argv) {
  char *rhost, *rport, *name;
  int sock, sock2;
  int xorkey, namelen, reconntime, daemonize;
  int reqport;
  unsigned char buf[8192];
  pid_t mypid;

  /* parse args (remote IP and port) */
  if (parseargv(argc, argv, &rhost, &rport, &xorkey, &name, &namelen, &reconntime, &daemonize) < 0) {
    printf("\nbackconn-cli version " pVer "\n\n"
           "Usage: backconn-client [options] rhost rport\n"
           "\n"
           "options:\n"
           "  -x key   set a xor key (1..255) to xor all communication\n"
           "  -n name  sets a name to be advertised to backconn-adm\n"
           "  -t sec   how long to wait between connection attempts (default: 5s)\n"
           "  -d       daemonize the process (switch into background)\n"
           "  -q       quiet mode: no console output\n"
           "  -s       log to syslog() additionally to console messages\n");
    return(1);
  }

  /* if quiet mode, close stdin, stdout and stderr */
  if (stderrflag == 0) {
    fclose(stdin);
    fclose(stdout);
    fclose(stderr);
  }

  /* daemonize, if requested */
  if (daemonize != 0) {
    mypid = fork();
    if (mypid > 0) return(0); /* parent dies and child takes over */
  }

  /* I do not care about my children, but I do not want them to become zombies */
  signal(SIGCHLD, SIG_IGN);

  for (;; sleep(reconntime)) { /* try connecting in a loop, waiting a moment between each retry */
   ReconnectWithoutWaiting:
    reqport = -1;
    /* try connecting to remote host, until it works */
    if (stderrflag != 0) printf("connecting to server %s:%s...\n", rhost, rport);
    sock = connect_to(rhost, rport);
    if (sock < 0) continue;
    if (stderrflag != 0) printf("connected to server %s:%s\n", rhost, rport);
    if (syslogflag != 0) syslog(LOG_INFO, "connected to server %s:%s\n", rhost, rport);

    /* send my name */
    buf[0] = namelen;
    memmove(buf+1, name, namelen);
    if (sendall(sock, buf, namelen+1) != 0) {
      if (stderrflag != 0) fprintf(stderr, "send() failure\n");
      if (syslogflag != 0) syslog(LOG_ERR, "send() failure");
      close(sock);
      continue;
    }

    /* wait for a two-bytes sequence or end of conn */
    for (;;) {
      if (recv(sock, buf, 1, 0) != 1) break; /* I could fetch 2 bytes at once, but then i could potentially get 1 byte only anyway, so it's easier to do it 1 by 1 */
      if (recv(sock, buf+1, 1, 0) != 1) break;
      xorstr(buf, xorkey, 2);
      reqport = buf[0];
      reqport <<= 8;
      reqport |= buf[1];
      if (reqport != 0) break;
      /* else it's just a keepalive - send back a 0xAA */
      buf[0] = 0xAA;
      xorstr(buf, xorkey, 1);
      if (send(sock, buf, 1, 0) != 1) {
        reqport = -1;
        break;
      }
    }
    /* if error, restart the whole thing */
    if (reqport < 0) {
      if (stderrflag != 0) fprintf(stderr, "connection with server broken\n");
      if (syslogflag != 0) syslog(LOG_ERR, "connection with server broken");
      close(sock);
      continue;
    }

    /* fork out a child, and let the parent restart */
    mypid = fork();
    if (mypid > 0) {  /* parent: close connection and restart without waiting */
      close(sock);
      goto ReconnectWithoutWaiting;
    }
    break; /* child: break out to handle the connection, and then die with it */
  }

  /* child: being here means I have to open the secondary connection to localhost */
  sprintf((char *)buf, "%d", reqport);
  sock2 = connect_to("127.0.0.1", (char *)buf);
  if (sock2 < 0) {
    if (stderrflag != 0) fprintf(stderr, "secondary connection to 127.0.0.1 on port %s failed.\n", buf);
    if (syslogflag != 0) syslog(LOG_ERR, "secondary connection to 127.0.0.1 on port %s failed.", buf);
    close(sock);
    return(0);
  }

  /* I have two connections now - let's just juggle data back and forth until something breaks */
  proxy(sock, sock2, buf, sizeof(buf), xorkey);
  close(sock);
  close(sock2);
  if (stderrflag != 0) fprintf(stderr, "connection ended\n");
  if (syslogflag != 0) syslog(LOG_INFO, "connection ended");

  return(0);
}
