/*
 * common funtions (used by both backconn-cli and backconn-adm)
 *
 * This file is part of the backconn project. http://backconn.sourceforge.net
 * Copyright (C) 2017, 2018 Mateusz Viste
 *
 * ===================== PUBLISHED UNDER THE MIT LICENSE =====================
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * ===========================================================================
 */

#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <string.h>      /* strerror() */
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>        /* time() */
#include <unistd.h>      /* NULL */


#include "common.h"

/* sends l bytes from buffer buf to socket s. return 0 on success, -1 otherwise */
int sendall(int s, const void *buf, int l) {
  int bytes, bytesdone = 0;
  for (;;) {
    bytes = send(s, ((unsigned char *)buf)+bytesdone, l, 0);
    if (bytes < 0) return(-1);
    l -= bytes;
    if (l == 0) return(0);
    bytesdone += bytes;
  }
}

/* a wrapper around recv, but with a controlled timeout (in seconds) */
int recvtimeout(int sock, void *bufptr, int len, int timeout, int flags) {
  int res = 0, recvbytes, stilltime;
  unsigned char *buff = bufptr;
  struct timeval t;
  fd_set fds;
  time_t starttime = time(NULL);
  time_t endtime;
  endtime = starttime + timeout;
  for (;;) {
    FD_ZERO(&fds);
    FD_SET(sock, &fds);
    t.tv_sec = 1;
    t.tv_usec = 0;
    select(sock+1, &fds, NULL, NULL, &t);
    if (FD_ISSET(sock, &fds)) {
      recvbytes = recv(sock, buff, len, MSG_DONTWAIT);
      if (recvbytes < 1) {
        if (res == 0) return(recvbytes);
        return(res);
      }
      res += recvbytes;
      buff += recvbytes;
      len -= recvbytes;
    }
    if (time(NULL) < endtime) {
      stilltime = 1;
    } else {
      stilltime = 0;
    }
    if ((flags & MSG_WAITALL) && (len > 0) && (stilltime)) continue;
    if (res > 0) return(res);
    if (stilltime) continue;
    return(EAGAIN);
  }
}

/* "encrypt" a string of bytes using a single-byte xor key */
void xorstr(void *buf, int xorkey, int len) {
  unsigned char *bufptr = buf;
  while (len--) bufptr[len] ^= xorkey;
}


/* proxifies data from one connection to another (and vice versa), effectively
 * juggling bytes back and forth. returns when one of the connections breaks.
 * this function does not close any of the connections - do cleanup yourself!
 * returns 0 if s1 breaks, 1 if s2 breaks, and -1 if some other error happens */
int proxy(int s1, int s2, unsigned char *buf, int buflen, int xorkey) {
  fd_set socks;
  struct timeval t;
  int rbytes;
  int highestsock;
  int s[2];
  int i;
  s[0] = s1;
  s[1] = s2;
  if (s2 > s1) {
    highestsock = s2;
  } else {
    highestsock = s1;
  }
  highestsock += 1;
  for (;;) {
    FD_ZERO(&socks);
    FD_SET(s1, &socks);
    FD_SET(s2, &socks);
    t.tv_sec = 10;
    t.tv_usec = 0;
    if (select(highestsock, &socks, NULL, NULL, &t) < 0) return(-1);
    /* check both sockets */
    for (i = 0; i < 2; i++) {
      if (FD_ISSET(s[i], &socks)) {
        rbytes = recv(s[i], buf, buflen, 0);
        if (rbytes <= 0) return(i);
        if ((xorkey != 0) && (rbytes > 0)) xorstr(buf, xorkey, rbytes);
        if (sendall(s[i^1], buf, rbytes)) return(i^1);
      }
    }
  }
}


/* enable TCP keepalive on socket */
void enable_tcpkeepalive(int sock) {
  int one = 1;
  int maxcnt = 3;
  int freq = 10;

  /* enable TCP keepalive on the socket */
  setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &one, sizeof(int));

  /* set the max keepalive count (how many missed means 'broken') */
  setsockopt(sock, SOL_TCP, TCP_KEEPCNT, &maxcnt, sizeof(int));

  /* keepalive idle time (when to start the keepalive) */
  setsockopt(sock, SOL_TCP, TCP_KEEPIDLE, &freq, sizeof(int));

  /* keepalive interval (how often to do it once started) */
  setsockopt(sock, SOL_TCP, TCP_KEEPINTVL, &freq, sizeof(int));
}
